"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.toSafe = toSafe;
exports.wrapGenerator = wrapGenerator;

var _primordials = require("./primordials.cjs");

/**
 * Wrap ArrayIterator If Array.prototype [@@iterator] has been modified
 *
 * @type {<T>(array: T[]) => Iterable<T>}
 */
function toSafe(array) {
  if (array[_primordials.SymbolIterator] === _primordials.NativeArrayPrototypeSymbolIterator) {
    return array;
  }

  const arrayIterator = (0, _primordials.ArrayPrototypeSymbolIterator)(array);
  return (0, _primordials.ObjectFreeze)(
  /** @type {any} */
  {
    next() {
      return (0, _primordials.ArrayIteratorPrototypeNext)(arrayIterator);
    },

    return: undefined,
    throw: undefined,

    [_primordials.SymbolIterator]() {
      return this;
    }

  });
}
/** @type {WeakMap<{}, Generator<any>>} */


const generators = new _primordials.NativeWeakMap();
/** @see https://tc39.es/ecma262/#sec-%arrayiteratorprototype%-object */

const DummyArrayIteratorPrototype = (0, _primordials.ObjectCreate)(_primordials.IteratorPrototype, {
  next: {
    value: function next() {
      const generator = (0, _primordials.WeakMapPrototypeGet)(generators, this);
      return (0, _primordials.GeneratorPrototypeNext)(generator);
    },
    writable: true,
    configurable: true
  },
  [_primordials.SymbolToStringTag]: {
    value: "Array Iterator",
    configurable: true
  }
});
/** @type {<T>(generator: Generator<T>) => IterableIterator<T>} */

function wrapGenerator(generator) {
  const dummy = (0, _primordials.ObjectCreate)(DummyArrayIteratorPrototype);
  (0, _primordials.WeakMapPrototypeSet)(generators, dummy, generator);
  return dummy;
}